/*
 * libtaskstats_snap - snapshot library using taskstats interface
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#ifndef _LIBTASKSTATS_SNAP_H_
#define _LIBTASKSTATS_SNAP_H_

#include <sys/time.h>
#include "taskstats_new.h"
#include "libtaskstats.h"

#ifdef __cplusplus
extern "C" {
#endif

struct ts_snap;
struct ts_snap_handle;
struct ts_snap_iter;

#define TS_SNAP_TASK_EXIT	1
#define TS_SNAP_TASK_FORK	2
#define TS_SNAP_TASK_RUN	4

/**
 * ts_snap_iter_new - Get iterator for snapshot
 * @tss:   Snapshot
 * @flags: Events to iterate (either TS_SNAP_TASK_EXIT, TS_SNAP_TASK_FORK or
 *         TS_SNAP_TASK_RUN)
 *
 * Returns snapshot iterator.
 */
extern struct ts_snap_iter *ts_snap_iter_new(struct ts_snap* tss, int flags);

/**
 * ts_snap_iter_free - Free snapshot iterator
 * @tss_i: Snapshot iterator to free
 */
extern void ts_snap_iter_free(struct ts_snap_iter *tss_i);

/**
 * ts_snap_iter_next - Return next taskstats from iterator
 * @tss_i: Snapshot iterator
 *
 * Returns taskstats.
 */
extern struct taskstats *ts_snap_iter_next(struct ts_snap_iter *tss_i);

extern unsigned long ts_snap_iter_cnt(struct ts_snap_iter *tss_i);

/**
 * ts_snap_init - Return new snap handle
 * @flags: Specify which events should be in the snapshot (e.g
 *         TS_SNAP_TASK_EXIT|TS_SNAP_TASK_RUN)
 *
 * Returns new snap handle.
 */
extern struct ts_snap_handle *ts_snap_init(int flags);

/**
 * ts_snap_exit - Free resources associated with snap handle
 * @tss_h: Snap handle to free
 */
extern void ts_snap_exit(struct ts_snap_handle *tss_h);

/**
 * ts_snap_next - Return next task snapshot
 * @tss_h: Snap handle
 *
 * The snapshot contains all running tasks and exit/fork events that
 * happened since the last snapshot that had been done via the
 * specified ts_snap_handle.
 */
extern struct ts_snap *ts_snap_next(struct ts_snap_handle *tss_h);

/**
 * ts_snap_free - Free resources associated with snapshot
 * @tss: Snapshot
 */
extern void ts_snap_free(struct ts_snap *tss);

/**
 * ts_snap_update_ns - Get update time of snapshot
 *
 * Returns update time stamp in nanoseconds.
 */
extern __u64 ts_snap_update_ns(struct ts_snap *tss);

#ifdef __cplusplus
}
#endif

#endif
