/*
 * libtaskstats - Library for easy access to taskstats interface
 *
 * Copyright IBM Corp. 2006-2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 *            Balbir Singh <balbir@linux.vnet.ibm.com>
 *            Shailabh Nagar
 */

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <poll.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/inotify.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <linux/net.h>
#include <linux/genetlink.h>
// #include <linux/cgroupstats.h>

#include <netlink/netlink.h>
#include <netlink/utils.h>
#include <netlink/netlink-kernel.h>
#include <netlink/msg.h>
#include <netlink/attr.h>

#include "libtaskstats.h"
#include "helper.h"

#define CMD_SEQ_NR 4711U

#ifdef NL_V1
#define __nl_recv nl_recv
#else
#define nl_handle_get_fd nl_socket_get_fd
static inline int __nl_recv(struct nl_handle *nlh, struct sockaddr_nl *sockaddr,
			    unsigned char **buf)
{
	return nl_recv(nlh, sockaddr, buf, NULL);
}
#endif

struct ts_socket {
	struct nl_handle	*nl_sock;
};

static struct {
	__u16		fam_id;
	unsigned int	taskstats_version_kern;
	unsigned int	taskstats_version_prog_compile;
	unsigned int	taskstats_size;
	int		cnt_max;
} l;

static struct ts_socket *ts_socket_alloc(void)
{
	struct ts_socket *tsh;

	tsh = malloc(sizeof(*tsh));
	tsh->nl_sock = nl_handle_alloc();

	return tsh;
}

static void ts_socket_free(struct ts_socket *tsh)
{
	nl_handle_destroy(tsh->nl_sock);
	free(tsh);
}

/*
 * Create a raw netlink socket and bind
 */
struct ts_socket *ts_socket_create(void)
{
	struct ts_socket *tsh;

	tsh = ts_socket_alloc();

	if (tsh) {
		nl_disable_sequence_check(tsh->nl_sock);
		// nl_disable_auto_ack(tsh->nl_sock);
		nl_connect(tsh->nl_sock, NETLINK_GENERIC);
	}
	return tsh;
}

void ts_socket_close(struct ts_socket *tsh)
{
	nl_close(tsh->nl_sock);
	ts_socket_free(tsh);
}

static int l_recv_timeout(struct nl_handle *nlh, unsigned char **buf,
			  int timeout_s)
{
	struct sockaddr_nl peer;
	struct timeval tv;
	fd_set nlhs;
	int sd;

	FD_ZERO(&nlhs);
	sd = nl_handle_get_fd(nlh);
	tv.tv_sec = timeout_s;
	tv.tv_usec = 0;
	FD_SET(sd, &nlhs);
	if (select(sd + 1, &nlhs, 0, 0, &tv) < 0)
		return -ETIME;

	/* go ahead and receive select returns >= 0 */
	return __nl_recv(nlh, &peer, buf);
}

/*
 * Probe the controller in genetlink to find the family id
 * for the TASKSTATS family
 */
static int get_family_id(struct nl_handle *nlh)
{
	struct nlmsghdr req = {
		.nlmsg_type = GENL_ID_CTRL,
	};
	struct genlmsghdr genlh;
	struct nlmsghdr *rep;
	struct nl_msg *msg;
	struct nlattr *nla;
	unsigned char *rmsg;
	int n, ret, len, id = -1;

	/* Get family name */
	msg = (struct nl_msg *)nlmsg_build(&req);
	/* Add genetlink header */
	genlh.cmd = CTRL_CMD_GETFAMILY;
	genlh.version = 0x1;
	nlmsg_append(msg, &genlh, GENL_HDRLEN, 0);
	/* Add attributes */
	ret = nla_put_string(msg, CTRL_ATTR_FAMILY_NAME, TASKSTATS_GENL_NAME);
	if (ret < 0)
		goto err;
#ifdef NL_V1
	nl_send_auto_complete(nlh, nlmsg_hdr(msg));
#else
	nl_send_auto_complete(nlh, msg);
#endif

	/* go ahead and receive select returns >= 0 */
	n = l_recv_timeout(nlh, &rmsg, 5);
	if (n < 0)
		ERR_EXIT("no response from netlink in 5 seconds\n");
	rep = (struct nlmsghdr *)rmsg;
	while (nlmsg_ok(rep, n)) {
		nla = nlmsg_attrdata(rep, GENL_HDRLEN);
		len = nlmsg_attrlen(rep, GENL_HDRLEN);
		if (nla_ok(nla, len)) {
			nla = nla_find(nla, len, CTRL_ATTR_FAMILY_ID);
			if (nla) {
				id = nla_get_u16(nla);
				goto done;
			}
		}
		rep = nlmsg_next(rep, &n);
	}
	ERR_EXIT("Kernel does not have taskstats support\n");
done:
	free(rmsg);
	nlmsg_free(msg);
	return id;
err:
	return -1;
}

static void taskstats_get(struct taskstats *taskstats, struct nlmsghdr *rep,
			  int n, __u32 *seq)
{
	struct nlattr *nla, *nnla;
	int len, rtid;

	/* printf("XXX seq: %i %i %i %i\n",
	       rep->nlmsg_len, n, rep->nlmsg_seq, rep->nlmsg_pid); */
	*seq = rep->nlmsg_seq;
	while (nlmsg_ok(rep, n)) {
		nla = nlmsg_attrdata(rep, GENL_HDRLEN);
		len = nlmsg_attrlen(rep, GENL_HDRLEN);
		if (nla_ok(nla, len)) {

			nnla = nla_find(nla, len, TASKSTATS_TYPE_AGGR_TGID);
			if (!nnla)
				nnla = nla_find(nla, len,
						TASKSTATS_TYPE_AGGR_PID);
			if (!nnla)
				break;
			/* It is a nested attribute, move down */
			nnla = nla_data(nnla);
			rtid = nla_get_u32(nnla);
			nnla = nla_find(nnla, len, TASKSTATS_TYPE_STATS);
			memcpy(taskstats, nla_data(nnla),
			       MIN((unsigned int) nla_len(nnla),
				   l.taskstats_size));
			return;
		} else {
			break;
		}
		rep = nlmsg_next(rep, &n);
	}
}

int ts_read_multi(struct ts_socket *tsh, int cnt,
		  struct taskstats *ts_vec)
{
	struct sockaddr_nl peer;
	struct nlmsghdr *rmsg;
	__u32 seq;
	int i, n;

	for (i = 0; i < cnt; i++) {
next:
		n = __nl_recv(tsh->nl_sock, &peer, (unsigned char **) &rmsg);
		if (rmsg->nlmsg_type == NLMSG_ERROR) {
			struct nlmsgerr *err = NLMSG_DATA(rmsg);
			// printf("XXX: Error: %d\n", err->error);
			if (err->error == 0) {	// ACK message
				free(rmsg);
				goto next;
			}
			free(rmsg);
			return 0;
		}
		taskstats_get(&ts_vec[i], rmsg, n, &seq);
		if (seq != CMD_SEQ_NR + i)
			ERR_EXIT("Invalid sequence number: %i (should be %i)\n",
				 seq, CMD_SEQ_NR + i);
		if (ts_vec[i].ac_comm[0] == 0) {
			free(rmsg);
			return i;
		}
		free(rmsg);
	}
	return i;
}

int __nl_wait_for_ack(struct nl_handle *nlh)
{
	struct sockaddr_nl peer;
	struct nlmsghdr *rmsg;
	int n, rc = -1;

	n = __nl_recv(nlh, &peer, (unsigned char **) &rmsg);
	if (rmsg->nlmsg_type == NLMSG_ERROR) {
		struct nlmsgerr *err = NLMSG_DATA(rmsg);
		if (err->error == 0) {
			rc = 0;
			goto out;
		}
	}
out:
	free(rmsg);
	return rc;
}

#if 0
static int send_cmd(struct nl_handle *nlh, __u16 nla_type, void *nla_data,
		    int nla_len)
{
	struct genlmsghdr genlh;
	struct nl_data *nl_data;
	struct nlmsghdr req;
	struct nl_msg *msg;

	nl_data = nl_data_alloc(nla_data, nla_len); // XXX Free

	memset(&req, 0, sizeof(req));
	memset(&genlh, 0, sizeof(genlh));

	req.nlmsg_type = l.fam_id;
	req.nlmsg_flags = NLM_F_REQUEST;
	// req.nlmsg_pid = 0; // XXX tid;

	genlh.cmd = TASKSTATS_CMD_GET;
	genlh.version = 0x1; // XXX TASKSTATS_VERSION; // XXX 1?

	msg = (struct nl_msg *) nlmsg_build(&req);
	nlmsg_append(msg, &genlh, GENL_HDRLEN, 0);

	if (nla_put_data(msg, nla_type, nl_data) < 0)
		return -1;
	if (nl_send_auto_complete(nlh, msg) < 0)
		return -1;
	// if (nl_wait_for_ack(nlh) < 0)
	//	return -1;
	// XXX nlmsg_free(msg);
	return 0;
}
#else
static int send_cmd(struct nl_handle *nlh, __u16 nla_type, void *nla_data,
		    int nla_len)
{
	struct genlmsghdr genlh;
	struct nl_data *nl_data;
	struct nlmsghdr req;
	struct nl_msg *msg;
	int rc = 0;

	nl_data = nl_data_alloc(nla_data, nla_len);

	memset(&req, 0, sizeof(req));
	memset(&genlh, 0, sizeof(genlh));

	req.nlmsg_type = l.fam_id;
	req.nlmsg_flags = NLM_F_REQUEST;
	req.nlmsg_seq = CMD_SEQ_NR;
#ifdef NL_V1
	req.nlmsg_pid = getpid(); // XXX
#else
	req.nlmsg_pid = nl_socket_get_local_port(nlh);
#endif

	genlh.cmd = TASKSTATS_CMD_GET;
	genlh.version = 0x1; // XXX TASKSTATS_VERSION; // XXX 1?

	msg = (struct nl_msg *) nlmsg_build(&req);
	nlmsg_append(msg, &genlh, GENL_HDRLEN, 0);

	if (nla_put_data(msg, nla_type, nl_data) < 0) {
		rc = -1;
		goto out_free;
	}
#ifdef NL_V1
	if (nl_send(nlh, nlmsg_hdr(msg)) < 0) {
#else
	if (nl_send(nlh, msg) < 0) {
#endif
		rc = -1;
		goto out_free;
	}

out_free:
	nl_data_free(nl_data);
	nlmsg_free(msg);
	return rc;
}

#endif

int ts_cmd_get_pid(struct ts_socket *tsh, __u32 pid,
		   struct taskstats *taskstats)
{
	int rc;

	rc = send_cmd(tsh->nl_sock, TASKSTATS_CMD_ATTR_PID, &pid, sizeof(pid));
	if (rc < 0)
		ERR_EXIT("get_pid: error pid: %i\n", pid);
	return ts_read(tsh, taskstats);
}

int __ts_cmd_get_pids(struct ts_socket *tsh, __u32 pid, int cnt, __u64 time_ns,
		      struct taskstats *ts_vec)
{
	struct taskstats_cmd_pids cmd_pids;
	int rc;

	cmd_pids.time_ns = time_ns;
	cmd_pids.cnt = cnt;
	cmd_pids.pid = pid;

	rc = send_cmd(tsh->nl_sock, TASKSTATS_CMD_ATTR_PIDS, &cmd_pids,
			sizeof(cmd_pids));
	if (rc)
		return rc;
	return ts_read_multi(tsh, cnt, ts_vec);
}

int ts_cmd_get_pids(struct ts_socket *tsh, __u32 pid, int cnt, __u64 time_ns,
		       __u64 *time_new_ns, struct taskstats *ts_vec)
{
	int cnt_cur = 0, rc, req_cnt;
	__u32 pid_cur = pid;

	do {
		req_cnt =  MIN(cnt - cnt_cur, l.cnt_max);
		rc = __ts_cmd_get_pids(tsh, pid_cur, req_cnt, time_ns,
				       &ts_vec[cnt_cur]);
		if (rc < 0)
			ERR_EXIT("ts_cmd_get_pids failed");
		cnt_cur += rc;
		if (rc < req_cnt)
			break;
		pid_cur = ts_vec[cnt_cur - 1].ac_pid + 1;
	} while (cnt_cur != cnt);
	if (cnt_cur != 0)
		*time_new_ns = ts_vec[cnt_cur - 1].time_ns;
	else
		*time_new_ns = time_ns;
	return cnt_cur;
}

#if 0
int ts_cmd_get_pids(struct ts_socket *tsh, __u32 pid, int cnt, __u64 time_ns,
		       __u64 *time_new_ns, struct taskstats *ts_vec)
{
	struct taskstats ts_vec_int[cnt];
	__u64 time_new_int_ns;
	int rc, i, j, k, cnt_rc;

	cnt_rc = ___ts_cmd_get_pids(tsh, pid, cnt, time_ns, &time_new_int_ns,
				ts_vec);

	for (i = 0; i < 2; i++) {
		rc = ___ts_cmd_get_pids(tsh, pid, cnt, time_new_int_ns,
					&time_new_int_ns, &ts_vec_int[0]);
		k = 0;
		// fprintf(stderr, "XXX: %i\n", rc);
		for (j = 0; j < rc; j++) {
			for (; k < cnt_rc; k++) {
				if (ts_vec[k].ac_pid == ts_vec_int[j].ac_pid) {
					memcpy(&ts_vec[k], &ts_vec_int[j],
					       sizeof(struct taskstats));
					break;
				}
			}
		}
	}
	*time_new_ns = time_new_int_ns;
	return cnt_rc;
}
#endif

#if 0
int ts_cmd_get_tgids(struct ts_socket *tsh, __u32 tgid, int cnt, __u64 time_ns,
		     struct taskstats *ts_vec)
{
	struct taskstats_cmd_pids cmd_pids;
	int rc;

	cmd_pids.time_ns = time_ns;
	cmd_pids.cnt = cnt;
	cmd_pids.pid = tgid;

	rc = send_cmd(tsh->nl_sock, TASKSTATS_CMD_ATTR_TGIDS, &cmd_pids,
		      sizeof(cmd_pids));
	if (rc)
		return rc;
	return ts_read_multi(tsh, cnt, ts_vec);
}
#endif

int ts_read(struct ts_socket *tsh, struct taskstats *taskstats)
{
	struct nlmsghdr *rmsg;
	int n, rc = 0;
	__u32 seq;

next:
	n = l_recv_timeout(tsh->nl_sock, (unsigned char **) &rmsg, 0);
	if (rmsg->nlmsg_type == NLMSG_ERROR) {
		struct nlmsgerr *err = NLMSG_DATA(rmsg);
		// printf("XXX: Error: %d\n", err->error);
		if (err->error == 0) {	// ACK message
			free(rmsg);
			goto next;
		}
		rc = -EIO;
		goto out_free;
	}
	taskstats_get(taskstats, (void *) rmsg, n, &seq);
out_free:
	free(rmsg);
	return rc;
}

#if 0
int ts_reply_get(struct ts_socket *tsh, struct ts_reply **tsr)
{
	struct sockaddr_nl peer;
	struct nlmsghdr *rmsg;
	int n;

next:
	n = __nl_recv(tsh->nl_sock, &peer, (unsigned char **) &rmsg);
	if (rmsg->nlmsg_type == NLMSG_ERROR) {
		struct nlmsgerr *err = NLMSG_DATA(rmsg);
		// printf("XXX: Error: %d\n", err->error);
		if (err->error == 0)	// ACK message
			goto next;
		return -EIO;
	}
	*tsr = rmsg;
	return 0;
}
#endif

int ts_exit_register(struct ts_socket *tsh, const char *cpumask)
{
	int rc;

	rc = send_cmd(tsh->nl_sock, TASKSTATS_CMD_ATTR_REGISTER_CPUMASK,
		      (char *) cpumask, strlen(cpumask) + 1);
	printf("Sent register cpumask, retval %d\n", rc);
	if (rc < 0)
		ERR_EXIT_ERRNO("error sending register cpumask\n");
	return 0;
}

int ts_snapshot_register(struct ts_socket *tsh, const char *dir)
{
	int rc;

	rc = send_cmd(tsh->nl_sock, TASKSTATS_CMD_ATTR_REGISTER_SNAPSHOT,
		      (char *) dir, strlen(dir) + 1);
	printf("Sent register snapshot, retval %d\n", rc);
	if (rc < 0)
		ERR_EXIT_ERRNO("error sending register snapshot\n");
	return 0;
}

int ts_exit_deregister(struct ts_socket *tsh, const char *cpumask)
{
	int rc;

	rc = send_cmd(tsh->nl_sock, TASKSTATS_CMD_ATTR_DEREGISTER_CPUMASK,
		      (char *) cpumask, strlen(cpumask) + 1);
	printf("Sent deregister cpumask, retval %d\n", rc);
	if (rc < 0)
		ERR_EXIT_ERRNO("error sending register cpumask\n");
	return 0;
}

int ts_socket_size(struct ts_socket *tsh)
{
	int bufsz;
	socklen_t size = sizeof(bufsz);
	int sd;

	sd = ts_socket_get_fd(tsh);

	if (getsockopt(sd, SOL_SOCKET, SO_RCVBUF, &bufsz, &size) < 0)
		ERR_EXIT("Unable to get socket rcv buf size\n");
	printf("socket buffer size: %d kb\n", bufsz/1024);
	bufsz = 1024 * 1024 * 1024;
	printf("setting to: %d kb\n", bufsz/1024);
	if (setsockopt(sd, SOL_SOCKET, SO_RCVBUF, &bufsz, sizeof(bufsz)) < 0)
		ERR_EXIT("Unable to set socket rcv buf size to %d\n", bufsz);
	if (getsockopt(sd, SOL_SOCKET, SO_RCVBUF, &bufsz, &size) < 0)
		ERR_EXIT("Unable to get socket rcv buf size\n");
	printf("socket buffer size: %d kb\n", bufsz/1024);
	return 0;
}

int ts_socket_get_fd(struct ts_socket *tsh)
{
	return nl_handle_get_fd(tsh->nl_sock);
}

int ts_version(void)
{
	return l.taskstats_version_kern;
}

int __ts_init(unsigned int version_compile, unsigned int taskstats_size)
{
	struct taskstats taskstats;
	struct ts_socket *tsh;

	l.cnt_max = 50; // XXX make this dynamic!!!
	l.taskstats_version_prog_compile = version_compile;
	l.taskstats_size = taskstats_size;

	tsh = ts_socket_create();

	ts_socket_size(tsh);
	l.fam_id = get_family_id(tsh->nl_sock);
	if (!l.fam_id)
		ERR_EXIT("Error getting family id, errno %d\n", errno);

	ts_cmd_get_pid(tsh, 1, &taskstats);
	l.taskstats_version_kern = taskstats.version;

	printf("Taskstats Version (kern): %i\n", taskstats.version);
	printf("Taskstats Version (prog): %i\n", version_compile);
	printf("Size of taskstats (lib).: %lu\n",
	       (unsigned long) sizeof(struct taskstats));
	printf("Size of taskstats (prog): %u\n", taskstats_size);
	printf("Family ID...............: %d\n", l.fam_id);
	ts_socket_close(tsh);

	return 0;
}

