/*
 * libtaskstats - Library for easy access to taskstats interface using /proc/taskstats
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <poll.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include "libtaskstats.h"
#include "helper.h"

struct ts_socket {
	int	fd;
};

static struct {
	unsigned int	taskstats_version_kern;
	unsigned int	taskstats_version_prog_compile;
	unsigned int	taskstats_size_prog;
	unsigned int	taskstats_size_kern;
} l;

static struct ts_socket *ts_socket_alloc(void)
{
	struct ts_socket *tsh;

	tsh = malloc(sizeof(*tsh));
	return tsh;
}

static void ts_socket_free(struct ts_socket *tsh)
{
	free(tsh);
}

/*
 * Create a raw netlink socket and bind
 */
struct ts_socket *ts_socket_create(void)
{
	struct ts_socket *tsh;

	tsh = ts_socket_alloc();

	if (tsh) {
		tsh->fd = open("/proc/taskstats", O_RDONLY);
		if (tsh->fd < 0)
			ERR_EXIT_ERRNO("open taskstats failed\n");
	}
	return tsh;
}

void ts_socket_close(struct ts_socket *tsh)
{
	close(tsh->fd);
	ts_socket_free(tsh);
}

int l_cmd_get_pids(struct ts_socket *tsh, __u32 pid, int cnt, __u64 time_ns,
		    __u64 *time_min_ns, __u64 *time_max_ns,
		    struct taskstats *ts_vec)
{
	struct taskstats_cmd_pids cmd_pids;
	struct taskstats *ts;
	void *buf, *usr_ptr, *buf_ptr; 
	unsigned int i, pid_cnt;
	int rc;

	buf = malloc(l.taskstats_size_kern * cnt);
	if (!buf)
		ERR_EXIT("Malloc failed\n");
	cmd_pids.pid = pid;
	cmd_pids.cnt = cnt;
	cmd_pids.time_ns = time_ns;

	if (ioctl(tsh->fd, TASKSTATS_CMD_ATTR_PIDS, &cmd_pids) < 0)
		ERR_EXIT_ERRNO("ioctl failed\n");
	// rc = read(tsh->fd, ts_vec, sizeof(struct taskstats) * cnt);
	rc = read(tsh->fd, buf, l.taskstats_size_kern * cnt);
	if (rc < 0)
		ERR_EXIT_ERRNO("read failed\n");

	*time_min_ns = -1ULL;
	*time_max_ns = 0;
	usr_ptr = ts_vec;
	buf_ptr = buf;
	pid_cnt = ((unsigned int) rc) / l.taskstats_size_kern;
	for (i = 0; i < pid_cnt; i++) {
		memcpy(usr_ptr, buf_ptr,
		       MIN(l.taskstats_size_kern, l.taskstats_size_prog));
		ts = (struct taskstats *) usr_ptr;
		if (*time_min_ns > ts->time_ns)
			*time_min_ns = ts->time_ns;
		if (*time_max_ns < ts->time_ns)
			*time_max_ns = ts->time_ns;
		usr_ptr += l.taskstats_size_prog;
		buf_ptr += l.taskstats_size_kern;
	}

	free(buf);
	return pid_cnt;
}

#define RETRY_CNT		1
#define TIME_DIFF_MAX_NS	10000

int ts_cmd_get_pids(struct ts_socket *tsh, __u32 pid, int cnt, __u64 time_ns,
		    __u64 *time_new_ns, struct taskstats *ts_vec)
{
	__u64 time_min_ns, time_max_ns;
	int pid_cnt, i;

	/*
	 * Loop until we have an acceptable time interval of all
	 * taskstats.
	 *
	 * A big time inteval can happen, e.g. if the process is stopped
	 * because the virtual CPU is taken away by the hypervisor.
	 */
	for (i = 0; i < RETRY_CNT; i++) {
		pid_cnt = l_cmd_get_pids(tsh, pid, cnt, time_ns, &time_min_ns,
					 &time_max_ns, ts_vec);
		if (time_max_ns - time_min_ns < TIME_DIFF_MAX_NS)
			break;
	}
	*time_new_ns = time_min_ns;
	return pid_cnt;
}

int ts_cmd_get_pid(struct ts_socket *tsh, __u32 pid, struct taskstats *ts)
{
	__u64 time_ns;

	if (ts_cmd_get_pids(tsh, pid, 1, 0, &time_ns, ts) == 1)
		return 0;
	return -ENODEV;
}

void l_init(void)
{
	struct taskstats_cmd_pids cmd_pids;
	struct taskstats taskstats;
	struct ts_socket *tsh;
	int rc;

	cmd_pids.pid = 1;
	cmd_pids.cnt = 1;
	cmd_pids.time_ns = 0;

	tsh = ts_socket_create();
	if (ioctl(tsh->fd, TASKSTATS_CMD_ATTR_PIDS, &cmd_pids) < 0)
		ERR_EXIT_ERRNO("ioctl failed\n");
	rc = read(tsh->fd, &taskstats, sizeof(struct taskstats));
	if (rc < 0)
		ERR_EXIT_ERRNO("read failed\n");
	ts_socket_close(tsh);

	l.taskstats_version_kern = taskstats.version;
	l.taskstats_size_kern = rc;
}

int ts_read(struct ts_socket *tsh, struct taskstats *taskstats)
{
	ERR_EXIT("ts_read not implemented\n");
	return -EINVAL;
}

int ts_exit_register(struct ts_socket *tsh, const char *cpumask)
{
	ERR_EXIT("ts_exit_register not implemented\n");
	return -EINVAL;
}

int ts_snapshot_register(struct ts_socket *tsh, const char *dir)
{
	ERR_EXIT("ts_snapshot_register not implemented\n");
	return -EINVAL;
}

int ts_exit_deregister(struct ts_socket *tsh, const char *cpumask)
{
	ERR_EXIT("ts_exit_deregister not implemented\n");
	return -EINVAL;
}

int ts_socket_size(struct ts_socket *tsh)
{
	ERR_EXIT("ts_socke_size not implemented\n");
	return -EINVAL;
}

int ts_socket_get_fd(struct ts_socket *tsh)
{
	ERR_EXIT("ts_socket_get_fd not implemented\n");
	return -EINVAL;
}

int ts_version(void)
{
	return l.taskstats_version_kern;
}

int __ts_init(unsigned int version_compile, unsigned int taskstats_size)
{
	l.taskstats_version_prog_compile = version_compile;
	l.taskstats_size_prog = taskstats_size;

	l_init();

	printf("Taskstats Version (kern): %i\n", l.taskstats_version_kern);
	printf("Taskstats Version (prog): %i\n", version_compile);
	printf("Size of taskstats (lib).: %lu\n",
	       (unsigned long) sizeof(struct taskstats));
	printf("Size of taskstats (prog): %u\n", l.taskstats_size_prog);
	printf("Size of taskstats (kern): %u\n", l.taskstats_size_kern);

	return 0;
}
